#!/usr/bin/perl

############################################################################
# About:            Push locale DASH masternode data (blockheight, version, protocolversion, errors) to DashCentral.org (masternode monitoring service)
# Written by:       DashCentral.org
# Created:          02/06/2017
# Version:			6
############################################################################

# install required perl modules: see readme.txt

# load modules
use LWP::UserAgent;
use Net::SSL;
use IO::Socket::SSL;
use JSON::RPC::Client;
use JSON qw( decode_json encode_json );
use File::Spec::Functions qw(rel2abs);
use File::Basename;
use Getopt::Long;

# initialize modules
our $jclient = new JSON::RPC::Client;
our $ua = LWP::UserAgent->new;

# load masternode configuration and get test masternode restart command
my $config_path = dirname(rel2abs($0)) . '/dashcentral.conf';
my $test_restart_address = '';
GetOptions ('config=s' => \$config_path, 'testrestart=s' => \$test_restart_address);

require $config_path;

# define global vars
our @collected_masternode_data;

############################################################################
# Connect to all configured masternodes and collect the blockheight, version, protocolversion and errors variables
############################################################################

our $updater_version = 6;
our $count_rpc_errors = 0;
our $rpc_errors;
our $configuration_errors;
our $count_configuration_errors = 0;

foreach $address (keys %masternodes) {

	# check if user configured daemon crash autorestart and if provided paths are correct
	my $daemon_autorestart = 'disabled';
	if (!exists $masternodes{$address}{'daemon_autorestart'}){
		# backwards compatibility for < v4 config files
		$masternodes{$address}{'daemon_autorestart'}='disabled';
	}

	if ($masternodes{$address}{'daemon_autorestart'} eq 'enabled') {

		# set default datadir in case user entered nothing
		if ($masternodes{$address}{'daemon_datadir'} eq '') {
			$masternodes{$address}{'daemon_datadir'} = get_user_homedir() . '/.dashcore';
		}

		if (substr($masternodes{$address}{'daemon_binary'}, -1) eq '/') {
			$configuration_errors .=  $address . ":\t\t" . "Configuration error - daemon binary path has to be entered without trailing slash. Please correct the binary path: " . $masternodes{$address}{'daemon_binary'}. "\n";
			$count_configuration_errors ++;
		} elsif (!-e $masternodes{$address}{'daemon_binary'}) {
			$configuration_errors .=  $address . ":\t\t" . "Configuration error - provided DASH daemon binary for autorestart on crash does not exist: " . $masternodes{$address}{'daemon_binary'}. "\n";
			$count_configuration_errors ++;
		} elsif (substr($masternodes{$address}{'daemon_datadir'}, -1) eq '/') {
			$configuration_errors .=  $address . ":\t\t" . "Configuration error - daemon datadir path has to be entered without trailing slash. Please correct the the datadir path: " . $masternodes{$address}{'daemon_datadir'}. "\n";
			$count_configuration_errors ++;
		} elsif (!-d $masternodes{$address}{'daemon_datadir'}) {
			$configuration_errors .=  $address . ":\t\t" . "Configuration error - provided DASH datadir for autorestart on crash does not exist: " . $masternodes{$address}{'daemon_datadir'}. "\n";
			$count_configuration_errors ++;
		} else {
			$daemon_autorestart = 'enabled';
		}
	}

  
  	# setup RCP connection to DASH coindaemon
	$jclient->ua->credentials(
		$masternodes{$address}{'rpc_host'} . ':' . $masternodes{$address}{'rpc_port'}, 'jsonrpc', $masternodes{$address}{'rpc_user'} => $masternodes{$address}{'rpc_password'}
	);

	my $uri = 'http://' . $masternodes{$address}{'rpc_host'} . ':' . $masternodes{$address}{'rpc_port'} . '/';
	my $obj = {
		method  => 'getinfo',
		params  => [],
	};
	
	# perform getinfo request
	my $res = $jclient->call( $uri, $obj );
	
	# make sure there is no error
	if ($res){
		if ($res->is_error) {
			# connection error occurred
			$count_rpc_errors ++;
			$rpc_errors .= "- RPC error while connection to masternode " . $masternodes{$address}{'rpc_host'} . ':' . $masternodes{$address}{'rpc_port'} . ': ' . $res->error_message . "\n";
			push (@collected_masternode_data, {'address' => $address, 'blockheight' => 0, 'version' => 0, 'protocolversion' => 0, 'errors' => '', 'daemon_connection' => 'error', 'daemon_autorestart' => $daemon_autorestart, 'updater_version' => $updater_version});
		} else {
			# everything ok, push obtained data to collective array
			my %data = %{$res->result};
			push (@collected_masternode_data, {'address' => $address, 'blockheight' => $data{'blocks'}, 'version' => $data{'version'}, 'protocolversion' => $data{'protocolversion'}, 'errors' => $data{'errors'}, 'daemon_connection' => 'ok', 'daemon_autorestart' => $daemon_autorestart, 'updater_version' => $updater_version});
		}
	} else {
		# connection error occurred
		$count_rpc_errors ++;
		$rpc_errors .=  "- RPC error while connection to masternode " . $masternodes{$address}{'rpc_host'} . ':' . $masternodes{$address}{'rpc_port'} . ': ' . $jclient->status_line . "\n";
		push (@collected_masternode_data, {'address' => $address, 'blockheight' => 0, 'version' => 0, 'protocolversion' => 0, 'errors' => '', 'daemon_connection' => 'error', 'daemon_autorestart' => $daemon_autorestart, 'updater_version' => $updater_version});
	}

}

############################################################################
# Check if critical configuration errors occurred that require the abort of this script
############################################################################

if ($count_configuration_errors > 0) {
	print_header();
	print $configuration_errors . "\n";
	exit;
}

############################################################################
# Check if user wants to test the restart function for a single masternode
############################################################################

if ($test_restart_address ne '') {

	print_header();

	if (!exists $masternodes{$test_restart_address}) {
		print 'Masternode "' . $test_restart_address . '" can not be found in configuration' . "\n";
		exit(1);
	} elsif ($masternodes{$test_restart_address}{'daemon_autorestart'} ne 'enabled') {
		print 'Autorestart is not enabled for masternode "' . $test_restart_address . '". This is a prerequisite to run the restart test!' . "\n";
		exit(1);
	} else {
		my ($old_pid, $new_pid) = restart_daemon($masternodes{$test_restart_address}{'daemon_binary'}, $masternodes{$test_restart_address}{'daemon_datadir'});

		if ($new_pid == 0 && $old_pid == 0) {
			# new daemon has not been started correctly
			print $test_restart_address . ":\t\t" . 'PROBLEM - No running DASH daemon process has been detected. New DASH deamon could NOT be started. Something is wrong with your configuration!' . "\n";
		} elsif ($new_pid == 0 && $old_pid > 0) {
			# new daemon has not been started correctly
			print $test_restart_address . ":\t\t" . 'PROBLEM - Running DASH daemon has been detected and has been killed. New DASH daemon could NOT be started. Something is wrong with your configuration! (old PID ' . $old_pid . ')' . "\n";
		} elsif ($old_pid == 0) {
			# no old pid has been found, started new process
			print $test_restart_address . ":\t\t" . 'POSSIBLE PROBLEM - No running DASH daemon process has been detected. DASH daemon has been started (new PID ' . $new_pid . ')' . "\n";
		} else {
			# old process has been killed and new process has been started
			print $test_restart_address . ":\t\t" . 'SUCCESS - Running DASH daemon has been detected and has been killed. New DASH daemon process has been started! (old PID ' . $old_pid . ' / new PID ' . $new_pid . ')' . "\n";
		}

	}

	exit(0);

}

############################################################################
# Push collected data to DashCentral.org API
############################################################################
 
# DashCentral.org API URL
my $server_endpoint = "https://www.dashcentral.org/index.php?page=api&action=updater&version=" . $updater_version . "&api_key=" . $settings{'api_key'};

# configure HTTP request
my $req = HTTP::Request->new(POST => $server_endpoint);
$req->header('content-type' => 'application/json');

# convert masternode data to JSON and add it to HTTP request body
$req->content(encode_json(\@collected_masternode_data));
 
# perform request
my $resp = $ua->request($req);
if ($resp->is_success) {
	# request was successfull
	my $server_response = decode_json($resp->decoded_content);
	my @results = @{$server_response->{'results'}};
	my $count_results = 0;

	# print results returned from the server

	print_header();

	for my $i (0 .. $#results) {
		$count_results ++;
		if ($results[$i]{'status'} eq 'ok') {
			print $results[$i]{'address'} . ":\t\t" . $results[$i]{'status'} . "\n";

			# restart DASH daemon if user configured and restart and DashCentral server determined, that the DASH daemon is unfunctional
			if ($results[$i]{'action'} eq 'restart' && $masternodes{$results[$i]{'address'}}{'daemon_autorestart'} eq 'enabled') {
				
				my ($old_pid, $new_pid) = restart_daemon($masternodes{$results[$i]{'address'}}{'daemon_binary'}, $masternodes{$results[$i]{'address'}}{'daemon_datadir'});

				if ($old_pid == 0) {
					# no old pid has been found, started new process
					print $results[$i]{'address'} . ":\t\t" . 'DASH daemon has crashed and been restarted (new PID ' . $new_pid . ')' . "\n";
				} else {
					# old process has been killed and new process has been started
					print $results[$i]{'address'} . ":\t\t" . 'DASH daemon is unfunctional or has hung up. Daemon has been restarted (old PID ' . $old_pid . ' / new PID ' . $new_pid . ')' . "\n";
				}
			}

		} else {
			print $results[$i]{'address'} . ":\t\t" . $results[$i]{'status'} . ' (' . $results[$i]{'message'} . ")\n";
		}
	}

	if ($count_results > 0) {
		print "\nUpdate status: " . $server_response->{'status'}  ."\n";
	} else {
		print "\nUpdate status: no masternodes configured in dashcentral.conf\n";
	}


	if ($count_rpc_errors > 0) {
		print "\n$count_rpc_errors RPC coind daemon connection error(s) occurred:\n";
		print $rpc_errors;
	}

} else {
	# hard HTTP connection error occurred
    print "Data POST to DashCentral failed with code: ", $resp->code, "\n";
    print "Data POST to DashCentral with message: ", $resp->message, "\n";
}

############################################################################
# Determine process id of dash daemon and restart it
############################################################################

sub restart_daemon {
	my $binary = shift;
	my $datadir = shift;

	my $old_pid = determine_daemon_pid($binary, $datadir);

	if ($old_pid > 0) {
		kill -9,$old_pid;
	}

	if ($datadir ne '') {
		system($binary . ' -datadir=' . $datadir);
	} else {
		system($binary);
	}

	my $new_pid = determine_daemon_pid($binary, $datadir);

	return ($old_pid, $new_pid);

}

# determine daemon process id
sub determine_daemon_pid {
	my $binary = shift;
	my $datadir = shift;

	my $user_homedir = get_user_homedir();

	my $ps_output = `ps -eo pid:1,args:1`;

	my $determined_pid = 0;

	foreach (split(/\n/,$ps_output)) {
		my $line = $_;

		my $regexp1 = '([0-9]+)\s' . $binary . '\s-datadir=' . $datadir;
		if ($line =~ m/$regexp1/isg) {
			# dash daemon has been started with appended datadir
			$determined_pid = $1;
			last;
		} else {	
			# if datadir has not been set or datadir equals user homedir, it's highly likely that dashd has been started without datadir command line appendix.
			# try go determine pid using the basis command structure
			if ($datadir eq '' || $datadir eq $user_homedir . '/.dashcore') {
				my $regexp2 = '([0-9]+)\s' . $binary;
				if ($line =~ m/$regexp2/isg) {
					$determined_pid = $1;
					last;
				}
			}

		}
	}

	return $determined_pid;
}

# determine user home directyl
sub get_user_homedir {
	my $user_homedir = '/root';
	if (exists $ENV{'HOME'}){
		$user_homedir = $ENV{'HOME'};
	}
	return $user_homedir;
}

# print header
sub print_header {
	print "############################################################\n# DashCentral server data transfer results (Version " . $updater_version . "):\n############################################################\n\n";
}
